@@ -1,5 +1,5 @@
 #!/usr/bin/env python
-#from __future__ import (absolute_import, division, print_function, unicode_literals) # TODO: Remove this?
+
 import subprocess
 import sys
 import os
@@ -14,9 +14,11 @@ import argparse
 import signal
 from textwrap import wrap
 from distutils.version import LooseVersion
+from packaging import version
 
 from subprocess import PIPE, Popen
 from threading  import Thread, Event, Timer
+import webbrowser
 
 from colorama import init
 init()
@@ -27,7 +29,7 @@ except ImportError:
     from queue import Queue, Empty  # python 3.x
 
 from .templates import slurm_server_script, slurm_batch_script, mem_script
-from .utils import execute, modpath, on_windows, str_to_mb
+from .utils import execute, modpath, on_windows, str_to_mb, seconds2string, human2walltime
 
 # global run event to communicate with threads
 RUN_EVENT = None
@@ -65,22 +67,18 @@ ON_POSIX = 'posix' in sys.builtin_module_names
 #             print(data, end="", file=sys.stderr)
 
 
-if sys.version_info < (3,0):
-    input = raw_input
-
-
 def check_for_conda_update():
     """Checks for a more recent conda version and prints a message.
     """
     cmd = 'conda search -c kaspermunch slurm-jupyter'
-    conda_search = subprocess.check_output(cmd.split()).decode()
-    newest_version = conda_search.strip().split('\n')[-1].split()[1]
+    conda_search = subprocess.check_output(cmd, shell=True).decode()
+    newest_version = conda_search.strip().splitlines()[-1].split()[1]
     cmd = 'conda list -f slurm-jupyter'
-    conda_search = subprocess.check_output(cmd.split()).decode()
-    this_version = conda_search.strip().split('\n')[-1].split()[1]
+    conda_search = subprocess.check_output(cmd, shell=True).decode()
+    this_version = conda_search.strip().splitlines()[-1].split()[1]
     if LooseVersion(newest_version) > LooseVersion(this_version):
         msg = '\nA newer version of slurm-jupyter exists ({}). To update run:\n'.format(newest_version)
-        msg += '\n\tconda update -c kaspermunch slurm-jupyter\n'
+        msg += '\n\tconda install -c kaspermunch "slurm-jupyter"\n'
         print(RED + msg + ENDC)
 
 
@@ -88,12 +86,13 @@ def check_for_conda_update():
 def kbintr_handler(signal, frame):
     """Intercepts KeyboardInterrupt and asks for confirmation.
     """
-    msg = BLUE+'\nAre you sure? y/n: '+ENDC
-    try:
-        if input(msg) == 'y':
-            raise KeyboardInterrupt
-    except RuntimeError: # in case user does Ctrl-C instead of y
-        raise KeyboardInterrupt
+    # msg = BLUE+'\nAre you sure? y/n: '+ENDC
+    # try:
+    #     if input(msg) == 'y':
+    #         raise KeyboardInterrupt
+    # except RuntimeError: # in case user does Ctrl-C instead of y
+    #     raise KeyboardInterrupt
+    raise KeyboardInterrupt
 
 
 def kbintr_repressor(signal, frame):
@@ -111,13 +110,14 @@ def get_cluster_uid(spec):
     Returns:
         int: User id.
     """
-    process = subprocess.Popen(
+    process = Popen(
         'ssh {user}@{frontend} id'.format(**spec),
         shell=True,
         universal_newlines=True,
         stdout=subprocess.PIPE,
         stderr=subprocess.PIPE)
     stdout, stderr = process.communicate()
+    assert not process.returncode
     uid = int(re.search('uid=(\d+)', stdout).group(1))
     return uid
 
@@ -133,20 +133,19 @@ def submit_slurm_server_job(spec, verbose=False):
         str: Slurm job id.
     """
 
-    cmd = 'ssh {user}@{frontend} cat - > {tmp_dir}/{tmp_script} ; mkdir -p {tmp_dir} ; {slurm} ; sbatch {tmp_dir}/{tmp_script} '.format(**spec)
+    cmd = 'ssh {user}@{frontend} cat - > {tmp_dir}/{tmp_script} ; mkdir -p {tmp_dir} ; sbatch {tmp_dir}/{tmp_script} '.format(**spec)
         
     if verbose: print("script ssh transfer:", cmd, sep='\n')
 
     script = slurm_server_script.format(**spec)
     if verbose: print("slurm script:", script, sep='\n')
 
-    if sys.version_info >= (3,0): script = script.encode()
+    script = script.encode()
     stdout, stderr = execute(cmd, stdin=script) # hangs untill submission
 
     # get stdour and stderr and get jobid
-    if sys.version_info >= (3,0):
-        stdout = stdout.decode()
-        stderr = stderr.decode()
+    stdout = stdout.decode()
+    stderr = stderr.decode()
     try:
         job_id = re.search('Submitted batch job (\d+)', stdout).group(1)
     except AttributeError:
@@ -173,22 +172,18 @@ def submit_slurm_batch_job(spec, verbose=False):
     script = slurm_batch_script.format(**spec)
     if verbose: print("slurm script:", script, sep='\n')
 
-    # if sys.version_info >= (3,0):
-    #     script = script.encode()
-
     tmp_script_path = "{tmp_dir}/{tmp_script}".format(**spec)
     with open(tmp_script_path, 'w') as f:
         f.write(script)
 
-    cmd = '{slurm} && sbatch {tmp_dir}/{tmp_script} '.format(**spec)
+    cmd = 'sbatch {tmp_dir}/{tmp_script} '.format(**spec)
     if verbose: print("command:", cmd, sep='\n')
    
     stdout, stderr = execute(cmd, shell=True) # hangs untill submission
 
     # get stdour and stderr and get jobid
-    if sys.version_info >= (3,0):
-        stdout = stdout.decode()
-        stderr = stderr.decode()
+    stdout = stdout.decode()
+    stderr = stderr.decode()
     try:
         job_id = re.search('Submitted batch job (\d+)', stdout).group(1)
     except AttributeError:
@@ -214,16 +209,15 @@ def wait_for_job_allocation(spec, verbose=False):
     # wait a bit to make sure jobinfo database is updated
     time.sleep(20)
 
-    cmd = 'ssh {user}@{frontend} {slurm} ; squeue --noheader --format %N -j {job_id}'.format(**spec)        
+    cmd = 'ssh {user}@{frontend} squeue --noheader --format %N -j {job_id}'.format(**spec)        
     stdout, stderr = execute(cmd)
-    if sys.version_info >= (3,0): stdout = stdout.decode()
+    stdout = stdout.decode()
     node_id = stdout.strip()
 
     while not node_id: #not m or m.group(1) == 'None':
         time.sleep(10)
         stdout, stderr = execute(cmd)
-        if sys.version_info >= (3,0):
-            stdout = stdout.decode()
+        stdout = stdout.decode()
         # m = regex.search(stdout)
         node_id = stdout.strip()
     if verbose: print(stdout)
@@ -255,7 +249,8 @@ def open_output_connection(cmd, spec):
     Returns:
         (subprocess.Popen, threading.Thread, Queue.Queue): Process, Thread and Queue.
     """      
-    p = Popen(cmd.split(), stdout=PIPE, stderr=PIPE, bufsize=0, close_fds=ON_POSIX)
+    # p = Popen(cmd.split(), stdout=PIPE, stderr=PIPE, bufsize=0, close_fds=ON_POSIX)
+    p = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE, bufsize=0, close_fds=ON_POSIX)
     q = Queue()
     t = Thread(target=enqueue_output, args=(p.stdout, q))
     t.daemon = True # thread dies with the program
@@ -273,7 +268,18 @@ def open_jupyter_stdout_connection(spec, verbose=False):
 
     Returns:
         (subprocess.Popen, threading.Thread, Queue.Queue): Process, Thread and Queue.
-    """       
+    """
+
+    file_created = False
+    cmd = 'ssh -q {user}@{frontend} [[ -f {tmp_dir}/{tmp_name}.{job_id}.out ]] && echo "File exists"'.format(**spec)
+    while not file_created:
+        if verbose: print("testing existence:", cmd)
+        stdout, stderr = execute(cmd)
+        if "File exists" in stdout.decode():
+            file_created = True
+        else:
+            time.sleep(10)
+
     cmd = 'ssh {user}@{frontend} tail -F -n +1 {tmp_dir}/{tmp_name}.{job_id}.out'.format(**spec)
     if verbose: print("jupyter stdout connection:", cmd)
     return open_output_connection(cmd, spec)
@@ -289,7 +295,18 @@ def open_jupyter_stderr_connection(spec, verbose=False):
 
     Returns:
         (subprocess.Popen, threading.Thread, Queue.Queue): Process, Thread and Queue.
-    """      
+    """
+
+    file_created = False
+    cmd = 'ssh -q {user}@{frontend} [[ -f {tmp_dir}/{tmp_name}.{job_id}.err ]] && echo "File exists"'.format(**spec)
+    while not file_created:
+        if verbose: print("testing existence:", cmd)
+        stdout, stderr = execute(cmd)
+        if "File exists" in stdout.decode():
+            file_created = True
+        else:
+            time.sleep(10)
+
     cmd = 'ssh {user}@{frontend} tail -F -n +1 {tmp_dir}/{tmp_name}.{job_id}.err'.format(**spec)
     if verbose: print("jupyter stderr connection:", cmd)
     return open_output_connection(cmd, spec)
@@ -327,6 +344,13 @@ def open_port(spec, verbose=False):
     # we have to set stdin=PIPE even though we eodn't use it because this
     # makes sure the process does not inherrit stdin from the parent process (this).
     # Otherwise signals are sent to the process and not to the python script
+
+    # TODO: Make some kind of check to make sure the port forwarding is working
+    # MAC:
+    # lsof -i -P | grep LISTEN 
+    # Linux:
+    # netstat -a | grep LISTEN
+
     port_q = Queue()
     port_t = Thread(target=enqueue_output, args=(port_p.stderr, port_q))
     port_t.daemon = True # thread dies with the program
@@ -334,20 +358,53 @@ def open_port(spec, verbose=False):
     return port_p, port_t, port_q
 
 
-def open_chrome(spec):
-    """Opens Chrome on localhost and port.
+def open_browser(spec, force_chrome=False):
+    """Opens default browser on localhost and port.
 
     Args:
         spec (dict): Parameter specification.
     """
-    if platform.platform().startswith('Darwin'):
+    if not spec['url']:
+        url = 'https://localhost:{port}'.format(**spec)
+    if platform.platform().startswith('Darwin') or platform.platform().startswith('macOS-'):
         chrome_path = 'open -a /Applications/Google\ Chrome.app %s'
+        if force_chrome and os.path.exists('/Applications/Google Chrome.app'):
+            webbrowser.get(chrome_path).open(url, new=2)
+        else:
+            webbrowser.open(spec['url'], new=2)
+    elif platform.platform().startswith('Windows'):
+        chrome_path = 'C:/Program Files (x86)/Google/Chrome/Application/chrome.exe %s'
+        if force_chrome and os.path.exists(os.path.abspath(os.path.join(os.sep, 'Program Files (x86)', 'Google', 'Chrome', 'Application', 'chrome.exe'))):
+            webbrowser.get(chrome_path).open(spec['url'], new=2)
+        else:
+            webbrowser.open(spec['url'], new=2)
     else:
         chrome_path = '/usr/bin/google-chrome %s'
-    url = 'https://localhost:{port}'.format(**spec)
-    print(url)
-    webbrowser.get(chrome_path).open(url)
+        if force_chrome and os.path.exists('/usr/bin/google-chrome'):
+            webbrowser.get(chrome_path).open(spec['url'], new=2)
+        else:
+            webbrowser.open(spec['url'], new=2)
 
+# TODO: make a check of jupyter lab version and give a user warning or abort if it is not 3
+def check_jupyterlab_version(spec):
+    """Check that jupyter lab version is >=3
+
+    Args:
+        spec (dict): Parameter specification.
+
+    Returns:
+        bool: whether version is 3
+    """
+    process = Popen(
+        'ssh {user}@{frontend} "conda activate simons_jupyter && conda list | grep grep \"jupyterlab \""'.format(**spec),
+        shell=True,
+        universal_newlines=True,
+        stdout=subprocess.PIPE,
+        stderr=subprocess.PIPE)
+    stdout, stderr = process.communicate()
+    assert not process.returncode
+    name, version, package = stdout.split()
+    return version.parse(version) >= version.parse("3.0.0")
 
 def transfer_memory_script(spec, verbose=False):
     """Transvers the a python script to the cluster, which monitors memory use on the node.
@@ -363,7 +420,7 @@ def transfer_memory_script(spec, verbose=False):
         
     if verbose: print("memory script:", script, sep='\n')
 
-    if sys.version_info >= (3,0): script = script.encode()
+    script = script.encode()
     stdout, stderr = execute(cmd, stdin=script) # hangs untill submission
 
 
@@ -476,6 +533,10 @@ def slurm_jupyter():
                     default=0.1,
                     type=float,
                     help="Time out in seconds for cross thread operations")
+    parser.add_argument("-C", "--chrome",
+                    dest="chrome",
+                    action='store_true',
+                    help="View in Chrome if available (ignoring default browser)")                    
     parser.add_argument("-v", "--verbose",
                     dest="verbose",
                     action='store_true',
@@ -487,6 +548,11 @@ def slurm_jupyter():
         print("Multiprocessign across multiple nodes not supported yet - sorry")
         sys.exit()
 
+    if args.time[-1] in 'smhdSMHD':
+        unit = args.time[-1].lower()
+        value = int(args.time[:-1])
+        args.time = human2walltime(**{unit:value})
+
     spec = {'user': args.user,
             'port': args.port,
             'environment': args.environment,
@@ -500,7 +566,6 @@ def slurm_jupyter():
             'total_memory': args.total_memory,
             'cwd': os.getcwd(),
             'sources_loaded': '',
-            'slurm': 'source /com/extra/slurm/14.03.0/load.sh', # TODO: is this requierd???
             'mem_script': 'mem_jupyter.py',
             'tmp_script': 'slurm_jupyter_{}.sh'.format(int(time.time())),
             'tmp_name': 'slurm_jupyter',
@@ -508,7 +573,8 @@ def slurm_jupyter():
             'frontend': args.frontend,
             'hostport': args.hostport,
             'job_name': args.name,
-            'job_id': None }
+            'job_id': None 
+            'url': None}
 
 
     # test ssh connection:
@@ -536,7 +602,7 @@ def slurm_jupyter():
         days, (hours, mins, secs) = 0, tup[0].split(':')
     else:
         days, (hours, mins, secs) = tup[0], tup[1].split(':')
-    end_time = time.time() + int(days) * 86400 + int(hours) * 3600 + int(mins) * 60 + int(secs)
+    end_time = int(time.time()) + int(days) * 86400 + int(hours) * 3600 + int(mins) * 60 + int(secs)
 
     if args.total_memory:
         spec['memory_spec'] = '#SBATCH --mem {}'.format(int(str_to_mb(args.total_memory)))
@@ -595,8 +661,7 @@ def slurm_jupyter():
                 except Empty:
                     break
                 else:
-                    if sys.version_info >= (3,0):
-                        line = line.decode()
+                    line = line.decode()
                     line = line.replace('\r', '\n')
                     print(line, end="")
             while True:
@@ -605,36 +670,48 @@ def slurm_jupyter():
                 except Empty:
                     break
                 else:
-                    if sys.version_info >= (3,0):
-                        mem_line = mem_line.decode()
-                    print(mem_line, end="")
+                    mem_line = mem_line.decode().rstrip() 
+                    secs_left = end_time - int(time.time())
+                    color = secs_left > 600 and BLUE or RED
+                    mem_line += '\t'+color+'Time: '+seconds2string(secs_left)+ENDC
+                    print(mem_line)
+
+                    # if secs_left <= 5*60:
+                    #     stdout, stderr = execute('ssh {user}@{frontend} pkill -9 -f "tail -F -n +1 {tmp_dir}/{tmp_name}.{job_id}"'.format(**spec))
+
+
             while True:
                 try:  
                     line = stderr_q.get(timeout=args.timeout)#get_nowait()
                 except Empty:
                     break
                 else:
-                    if sys.version_info >= (3,0):
-                        line = line.decode()
+                    line = line.decode()
                     line = line.replace('\r', '\n')
                     if 'SSLV3_ALERT_CERTIFICATE_UNKNOWN' not in line: # skip warnings about SSL certificate
                         print(line, end="")
 
-                    if re.search('Jupyter Notebook[\d. ]+is running', line):
+                    if re.search('Jupyter Notebook[\d. ]+is running', line) or re.search('Jupyter Server[\d. ]+is running', line):
                         port_p, port_t, port_q = open_port(spec, verbose=args.verbose)
 
-                        open_chrome(spec)
-                        print(BLUE+'If your browser says "Your connection is not private",',
-                        'click "Advanced" and then "Proceed etc. (unsafe)"\n',
-                        '(To skip this step in the future you can type this in the Chrome addres bar:\n',
-                        '"chrome://flags/#allow-insecure-localhost"',
-                        'and then select "Enable")'+ENDC)
+                    m = re.search('http://127.0.0.1:\d+/lab\?token=\S+', line)
+                    if m:
+                        spec['url'] = m.group(0)
+
+                        open_browser(spec, force_chrome=args.chrome)
+                        print(BLUE+' Your browser may complain that the connection is not private.\n',
+                                   'In Safari, you can proceed to allow this. In Chrome, you need"\n',
+                                   'to simply type the characters "thisisunsafe" while in the Chrome window.\n',
+                                   'Once ready, jupyter may ask for your cluster password.'+ENDC)
+
     except KeyboardInterrupt:
 
-        # not possible to do Keyboard interrupt from hereon out
+        # not possible to do Keyboard interrupt from here on out
         signal.signal(signal.SIGINT, kbintr_repressor)
 
-        # in try statements becuase these vars may not be defined at keyboard interrupt:
+        # TODO: Double Ctrl-C bypasses canceling of slurm job
+
+        # in try statements because these variabless may not be defined at keyboard interrupt:
         try:
             RUN_EVENT.clear()
             port_t.join()
@@ -663,7 +740,7 @@ def slurm_jupyter():
             pass
 
         print(BLUE+'\nCanceling slurm job running jupyter server'+ENDC)
-        stdout, stderr = execute('ssh {user}@{frontend} {slurm} ; scancel {job_id}'.format(**spec))
+        stdout, stderr = execute('ssh {user}@{frontend} scancel {job_id}'.format(**spec), check_failure=False)
         sys.exit()
 
 
@@ -689,6 +766,7 @@ def slurm_nb_run():
     parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                         description=description)
 
+    # TODO: make sure this only shares common  arguments
     add_slurm_arguments(parser)                  
 
     # nbconvert arguments
@@ -710,10 +788,10 @@ def slurm_nb_run():
                     dest="inplace",
                     action='store_true',
                     help="Output format.")
-    parser.add_argument("--cleanup",
-                    dest="cleanup",
-                    action='store_true',
-                    help="Removes un-executed notebooks generated using --spike and format other than 'notebook'")                           
+    # parser.add_argument("--cleanup",
+    #                 dest="cleanup",
+    #                 action='store_true',
+    #                 help="Removes un-executed notebooks generated using --spike and format other than 'notebook'")                           
 
     parser.add_argument("-s", "--spike",
                     dest="spike",
@@ -748,6 +826,11 @@ def slurm_nb_run():
 
     home = os.path.expanduser("~")
 
+    if args.time[-1] in 'smhdSMHD':
+        unit = args.time[-1].lower()
+        value = int(args.time[:-1])
+        args.time = human2walltime(**{unit:value})
+        
     spec = {'environment': args.environment,
             'walltime': args.time,
             'account': args.account,
@@ -812,10 +895,79 @@ def slurm_nb_run():
 
     import nbformat
 
+
+    # TODO: submit jobs trough GWF Do the whole thing in a tmp dir and only cp
+    # back to the destination as last command so that we do not get notebooks
+    # that are not run
+
+    # make a template that runs one notebook
+
+    from gwf import Workflow, AnonymousTarget
+
+    gwf = Workflow(defaults={'account': args.account})
+
+    def run_workflow(workflow):
+        from gwf.core import Backend, Graph, Scheduler
+        from gwf.conf import config
+
+        graph = Graph.from_targets(workflow.targets)
+        backend_cls = Backend.from_name('slurm')
+
+        with backend_cls() as backend:
+            scheduler = Scheduler(graph=graph, backend=backend, dry_run=False)
+            scheduler.schedule_many(list(graph))
+
+    def nbconvert(notebook_file_name, dependencies=[], inplace=False, 
+                  output_format='notebook', allow_errors=False, timeout=-1):
+        inputs = dependencies
+        outputs = [notebook_file_name]
+        options = {
+            'cores': args.cores,
+            'memory': args.total_memory,
+        }
+
+        if allow_errors:
+            allow_errors = '--allow-errors'
+        else:
+            allow_errors = ''
+
+        if inplace:
+            inplace = '--inplace'
+        else:
+            inplace = ''
+
+        spec = f'''
+        cp {notebook_file_name} $TMPDIR/`basename {notebook_file_name}` && \
+        nbconvert_cmd = "jupyter nbconvert --ClearOutputPreprocessor.enabled=True \
+            --ExecutePreprocessor.timeout={timeout} {allow_errors} {inplace and '--allow-errors' or ''} \
+                --to {output_format} --execute {notebook_file_name}" && \
+        cp $TMPDIR/`basename {notebook_file_name}` {notebook_file_name}
+        '''
+
+        return AnonymousTarget(inputs=inputs, outputs=outputs, options=options, spec=spec)
+
+
+
+
     if not args.spike:
-        command_list = [nbconvert_cmd.format(notebook=notebook, **spec) for notebook in notebook_list]
-        spec['commands'] = ' && '.join(command_list)
-        submit_slurm_batch_job(spec, verbose=args.verbose)
+        dependencies = []
+        for notebook in notebook_list:
+            gwf.target_from_template(
+                name=modpath(notebook, parent=''),
+                template=nbconvert(
+                    notebook_file_name=notebook,
+                    dependencies=dependencies,
+                    inplace=True,
+                    allow_errors=False, 
+                    timeout=args.timeout,
+                    output_format=args.format
+                )
+            )
+            dependencies.append(notebook)
+
+        # command_list = [nbconvert_cmd.format(notebook=notebook, **spec) for notebook in notebook_list]
+        # spec['commands'] = ' && '.join(command_list)
+        # submit_slurm_batch_job(spec, verbose=args.verbose)
 
     else:
         for spike_file in args.spike:
@@ -827,6 +979,9 @@ def slurm_nb_run():
 
             for notebook_path in notebook_list:
 
+                # TODO: instead add gwf targets with dependencies
+                # TODO: disallow --allow-errors when running more than one notebook
+
                 # find cell with %run magic if any
                 with open(notebook_path) as f:
                     nb = nbformat.read(f, as_version=4)
@@ -863,11 +1018,31 @@ def slurm_nb_run():
 
                 new_notebook_list.append(new_notebook_path)
 
-            if args.format != 'notebook' and args.cleanup:
-                param_nbconvert_cmd = nbconvert_cmd + ' && rm -f {notebook}'
-            else:
-                param_nbconvert_cmd = nbconvert_cmd
+            ###################################
+            dependencies = [spike_file]
+            for notebook in notebook_list:
+                gwf.target_from_template(
+                    name=modpath(notebook, parent=''),
+                    template=nbconvert(
+                        notebook_file_name=notebook,
+                        dependencies=dependencies,
+                        inplace=True,
+                        allow_errors=False, 
+                        timeout=args.timeout,
+                        output_format=args.format
+                    )
+                )
+                dependencies.append(notebook)
 
-            command_list = [param_nbconvert_cmd.format(notebook=notebook, **spec) for notebook in new_notebook_list]
-            spec['commands'] = ' && '.join(command_list)
-            submit_slurm_batch_job(spec, verbose=args.verbose)
+            ###################################
+
+            # if args.format != 'notebook' and args.cleanup:
+            #     param_nbconvert_cmd = nbconvert_cmd + ' && rm -f {notebook}'
+            # else:
+            #     param_nbconvert_cmd = nbconvert_cmd
+
+            # command_list = [param_nbconvert_cmd.format(notebook=notebook, **spec) for notebook in new_notebook_list]
+            # spec['commands'] = ' && '.join(command_list)
+            # submit_slurm_batch_job(spec, verbose=args.verbose)
+
+    run_workflow(gwf)
